"""Utility classes to build, install, and setup OpenAFS."""

from afsutil.__version__ import VERSION as __version__
from afsutil.__resources__ import resources
__version__   # hush lint
resources     # hush lint
